#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright (c) 2022 European Union.
#
# The tool was developed with the contribution of the Joint Research Centre of the European Commission.
#
# This program is free software: you can redistribute it and/or modify it under the terms of the European Union Public
# Licence, either version 1.2 of the License, or (at your option) any later version.
# You may not use this work except in compliance with the Licence.
#
# You may obtain a copy of the Licence at: https://joinup.ec.europa.eu/collection/eupl/eupl-guidelines-faq-infographics
#
# Unless required by applicable law or agreed to in writing, software distributed under the Licence is distributed on
# an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations under the Licence.


"""
SYNOPSIS
    yield2crop.py -i <folder_path> -o <folder_path> -y <year> [-v,--verbose] [-h,--help] [--version]

DESCRIPTION
    This snippet is for preprocessing different CAPRI maps (one per crop type) and convert it into one
    single croptype map to be used as input for crop provision (and crop pollination) accounting workflows

PREREQUISITES
    Python => 3.7.0

EXAMPLES
    python -m inca.cropprovision.preprocessing_yield2crop -i <folder_path> -o <folder_path> -y <year>
    OR
    python3 preprocessing_yield2crop.py -i <folder_path> -o <folder_path> -y <year>

EXIT STATUS
    Ctrl-C  =  cancel the script at any time

AUTHOR
    Bruno Smets <bruno.smets@vito.be>
"""

################## Importing modules ##################
# standard
import logging
import optparse
import os
import time
import subprocess
from datetime import datetime, timedelta
import random

# extra official modules
import rasterio
import numpy as np
from osgeo import gdal
import xml.etree.ElementTree as ET
from tqdm import tqdm

# own modules
import inca
import inca.common.geoprocessing as geo
from inca.common.geoprocessing import RasterType, number_blocks, block_window_generator, statistics_byArea

BLOCKSIZE = (100,100)
#BLOCKSIZE = (2,3)
test= ''  #'_test'

############## END importing modules ##################

logger = logging.getLogger(__name__)


############## MAIN ###################################
def main():
    start_time = time.time()
    # parse the arguments
    options = parse_args()

    # ini the logging
    inca.set_up_console_logging(options.verbose)

    # change umask driver
    try:
        old_umask = os.umask(0o000)
    except Exception:
        logger.exception('Could not change UMASK setting')
        pass

    logger.info('START OF MODULE: INCA YIELD2CROP map preprocessing')

    # init the metadata and raster extent of project
    metadata = geo.Metadata("Bruno Smets for INCA project",
                            "crop provision pre-processing"
                            )
    os.makedirs(os.path.normpath(options.path_out), exist_ok=True)

    # run the total lai generation
    logger.info('generating the crop map from CAPRI inputs at original resolution')
    year = options.pYear
    try:
        myVRT, offset_grass = build_vrt_capri(options, year)
        processing_capri(options, myVRT, offset_grass, year, metadata)
    except Exception:
        logger.exception('Error during processing of the datasets to generate the crop type map from CAPRI data')
        raise

    logger.info('All processing successfully done. Have a nice day.')
    # set umask to old settings
    try:
        os.umask(old_umask)
    except Exception:
        logger.exception('!!! Could not change UMASK setting back!!!')
        pass

    logger.debug('---------------------------')
    logger.debug('TOTAL TIME IN MINUTES: %10.4f', (time.time() - start_time) / 60)
    logger.debug('END OF MODULE: INCA crop provision YIELD2CROP preprocessing')


def parse_args():
    parser = optparse.OptionParser(formatter=optparse.TitledHelpFormatter(), usage=globals()['__doc__'],
                                   version="%prog v1.0")
    parser.add_option('-v', '--verbose', action='store_true', default=False, help='verbose output')

    parser.add_option('-i', '--path_in', help='Path to folder with CAPRI input data. Needed.', dest='path_in')
    parser.add_option('-o', '--path_out', help='Path to output folder. Needed.', dest='path_out')
    parser.add_option('-y', '--year', help='Processing year. Needed.', dest='pYear')

    # parse the given system arguments
    (options, args) = parser.parse_args()
    # do checks on the parsed options
    if (not options.path_in) or (not os.path.isdir(os.path.normpath(options.path_in))):
        parser.error("the -i argument for CAPRI input data is missing or the given path doesn't exist!!!")
    if (not options.pYear) or (not options.pYear.isdigit()):
        parser.error("the -y argument must be a year.")

    if len(args) != 0:
        parser.error('too many arguments')

    return options

def build_vrt_capri(options, year):

    my_vrt = os.path.join(options.path_out, 'CAPRI_' + str(year)+ test+ '.vrt')
    if os.path.exists(my_vrt):
        return my_vrt, 37  #TODO get from VRT

    # we do not care about memory so we run all in one chunk
    # get list of all files in input folder
    logger.info('Reading input files at location {}.'.format(os.path.normpath(options.path_in)))
    lFiles = []
    lFilesGrass = []
    for entry in os.scandir(os.path.normpath(options.path_in)):
        if entry.is_file() and str(options.pYear) in entry.name and entry.name.endswith(".tif"):
            if 'UAAR' in entry.name:
                logger.warning('Skip file {}'.format(entry.name))
                continue
            if ('GRAE' in entry.name) or ('GRAI' in entry.name):
                logger.warning('Delay file {} to append to end'.format(entry.name))
                lFilesGrass.append(entry.path)
                continue
            lFiles.append(entry.path)

    # push grassland crop types to end to enable easier distribution between cropland and grassland
    offset_grassland = len(lFiles)
    for f in lFilesGrass:
        lFiles.append(f)

    bandnames = []
    for f in lFiles:
        bandnames.append(os.path.split(f)[1][0:4])

    # create VRT file from all input crop individual files
    logger.info('Creating VRT file from all CAPRI input files')
    vrt_options = gdal.BuildVRTOptions(separate=True)
    gdal.BuildVRT(my_vrt, lFiles, options=vrt_options)

    # set bandnames in VRT file
    tree = ET.parse(my_vrt)
    root = tree.getroot()
    i = 0
    for band in root.iter("VRTRasterBand"):
        source = band.find("ComplexSource")
        file_name = source.find("SourceFilename").text

        # Description appears more useful though
        description = ET.SubElement(band, "Description")
        description.text = bandnames[i]
        i = i + 1
    # Update the file on disk
    tree.write(my_vrt)

    return my_vrt, offset_grassland

def distribute_crops(aET, aCrop, crops, off_grass, step):

    # distribute crop types according to contribution per ha
    aOut = np.zeros(aET.shape)
    # create maks for cropland and grassland pixels
    mCropland = (aET == 2)
    mGrassland = (aET == 3)
    mTotal = np.bitwise_or(mCropland, mGrassland)
    aOut = np.ma.masked_array(aOut, mask=~mTotal)
    #print('Distrute {} cropland and {} grassland, totals {} ha'.format(mCropland.sum(), mGrassland.sum(), aOut.count()))

    # loop over pixels to allocate croptypes to ET (cropland or grassland)
    try:
        for y in range(int(aET.shape[0] / step)):
            for x in range(int(aET.shape[1] / step)):
                pOut = aOut[y * step:(y * step) + step, x * step:(x * step) + step]
                mpCrop = mCropland[y * step:(y * step) + step, x * step:(x * step) + step]
                mpGrass = mGrassland[y * step:(y * step) + step, x * step:(x * step) + step]
                et_area = pOut.count()
                et_area_crop = pOut[mpCrop].count()
                crops_area_crop = aCrop[:off_grass, y, x].sum()
                et_area_grass = pOut[mpGrass].count()
                crops_area_grass = aCrop[off_grass:,y,x].sum()
                #print('Pixel {},{} has ET {} ha and Crop {} ha'.format(x,y,et_area,crops_area_crop))
                #print('ET location y {}:{} x {}:{}'.format(y * step, (y * step) + step, x * step,(x * step) + step))
                if ((np.ma.is_masked(crops_area_crop)) and (np.ma.is_masked(crops_area_grass))) \
                        or ((crops_area_crop == 0) and (crops_area_grass == 0)) or (et_area == 0):
                    #print('Crops area {}/{} or ET area {} is zero, skip'.format(crops_area_crop, crops_area_grass, et_area))
                    continue
                if (not et_area_crop == 0) and (crops_area_crop > et_area_crop):
                    #print('Warning: pixel {},{} has ET {} ha and Crop {} ha'.format(x, y, et_area_crop,crops_area_crop))
                    # apply ratio
                    aCrop[:off_grass, y, x] = aCrop[:off_grass, y, x] * et_area_crop / crops_area_crop
                if (not et_area_grass == 0) and (crops_area_grass > et_area_grass):
                    #print('Warning: pixel {},{} has ET {} ha and Crop-grass {} ha'.format(x, y, et_area_grass,crops_area_crop))
                    # apply ratio
                    aCrop[off_grass:, y, x] = aCrop[off_grass:, y, x] * et_area_grass / crops_area_grass
                # determine available positions to map crops
                lidx_crop = np.where(mpCrop)
                lidx_grass = np.where(mpGrass)
                idx_crop = np.vstack((lidx_crop[0], lidx_crop[1])).T.tolist()
                idx_grass = np.vstack((lidx_grass[0], lidx_grass[1])).T.tolist()

                #now loop over croptypes
                for i in range(crops):
                    #skip if no crop for this croptype for this pixel
                    crop_area = int(np.ceil(np.round(aCrop[i, y, x],1))) # roundup if one decimal
                    #if crop_area > int(np.round(aCrop[i,y,x])):
                    #        a=0
                    if crop_area == 0:
                        #print('Warning: skip crop {} as area {} too small'.format(i, aCrop[i,x,y]))
                        continue

                    #check if we're mapping on cropland or grassland
                    if i < off_grass:
                        #handle cropland
                        if (len(idx_crop) == 0) or np.ma.is_masked(aCrop[i, y, x]) or (aCrop[i, y, x] == 0): continue
                        # map crop area to pixels,  randomly map + mask mapped
                        #print('Mapping {} ha for crop-crop {}, left {}'.format(crop_area, i, len(idx_crop)))
                        for j in range(crop_area):
                            if len(idx_crop) <= 0:
                                # print('ERROR: no crop_area left')
                                break
                            pos = random.choice(idx_crop)
                            pOut[pos[0], pos[1]] = 101 + i  # crop is mapped from 101
                            # remove entry from idx
                            idx_crop.remove(pos)
                    else:
                        #handle grassland
                        if (len(idx_grass) == 0) or np.ma.is_masked(aCrop[i, y, x]) or (aCrop[i, y, x] == 0): continue
                        # map crop area to pixels,  randomly map + mask mapped
                        #print('Mapping {} ha for grass-crop {}, left {}'.format(crop_area, i, len(idx_grass)))
                        for j in range(crop_area):
                            if len(idx_grass) <= 0:
                                # print('ERROR: no crop_area left')
                                break
                            pos = random.choice(idx_grass)
                            pOut[pos[0], pos[1]] = 101 + i  # crop is mapped from 101
                            # remove entry from idx
                            idx_grass.remove(pos)

                #join GRAE and GRAI into one croptype class to represent J000
                #if len(pOut[pOut==(100+off_grass)]) != 0:
                #    a=0
                #pOut[pOut==(100+off_grass)] = 138
                # aOut[x * 10 + x:(x * 10) + 10, y * 10 + y:(y * 10) + 10] = pOut
        return aOut

    except Exception as err:
        print('Unexpected {}'.format(err))
        raise err

def processing_capri(options, my_vrt, offset_grass, year, metadata, scale_factor = 0.01, valid_range = [0, 32767]):
    ''' Join the CAPRI maps per crop into one single crop type map.

    NOTE: there is done no check - all tif files in folder are used to calculate the final crop type map.

    :param options: options
    :param metadata: metadata object from geoprocessing that needs be infused
    :param scale_factor: scale factor
    :param valid_range: valid range
    '''

    #VRT contains one band per croptype, expressed in hectares for 1kmx1km pixel (max 100ha per pixel)
    logger.info('Merging individual files in single croptype map.')
    base_path = options.path_out
    #my_ET = os.path.join(base_path,'CLC{}ACC_V2018_20_CAPRI.tif'.format(year))
    #my_ET = os.path.join(base_path,'ecosystem_types_L1_{}_CAPRI{}.tif'.format(year,test))
    my_ET = os.path.join('/data/inca_vol1/test/Bruno/CROP/CAPRI_map','ecosystem_types_L1_{}_CAPRI{}.tif'.format(year,test))
    my_CropOut = os.path.join(base_path,'CAPRI_{}_crops{}.tif'.format(year,test))

    with rasterio.open(my_vrt,'r') as src_crop, rasterio.open(my_ET) as src_et:
        crop_profile = src_crop.profile
        crops = crop_profile['count']
        dst_profile = src_et.profile

        # check if both maps align
        step = 10  #number of pixels in ET per crop pixel, in X or Y direction
        if (int(crop_profile['transform'][2]) != int(dst_profile['transform'][2])) or (int(crop_profile['transform'][5]) != int(dst_profile['transform'][5])):
            print('Crop map and Ecosystem map do not align, check coordinates')
            raise
        if int(crop_profile['transform'][0])/step != int(dst_profile['transform'][0]):
            print('Crop map and Ecosystem map do not align, check stepsize')
            raise

        # number of to process blocks
        nblocks = number_blocks(crop_profile, BLOCKSIZE)

        dst_profile.update(driver='GTiff',count=1, nodata=0, compress='lzw')

        with rasterio.open(my_CropOut, 'w', **dst_profile) as dst:
            t = tqdm(total=int(nblocks))

            # iterate over the blocks
            for block, window in block_window_generator(BLOCKSIZE, crop_profile['height'], crop_profile['width']):
                #print('** processing block {} with crop window {} **'.format(block, window))
                if (block[0]==43) and (block[1]==55):
                    b=0
                # read data
                ETwindow = ((window[0][0]*step,window[0][1]*step),(window[1][0]*step,window[1][1]*step))
                #print(' * read window ET {} for crop {}'.format(ETwindow, window))
                aET = src_et.read(1, window=ETwindow, masked=True)
                aCrop = src_crop.read(window=window, masked=True)  #multiple bands and once

                if aCrop.mask.all():
                    aSum= aCrop.sum(axis=0)
                    t.update()#all masked values
                    continue
                aSum= aCrop.sum(axis=0)

                aOut = distribute_crops(aET, aCrop, crops, offset_grass, step)

                # write to disk
                dst.write(aOut.astype(np.uint8), 1, window=ETwindow)
                t.update()

            t.close()
    a=0

if __name__ == '__main__':
    # hack to get the proper logger when running the script directly
    logger = logging.getLogger('inca.croppprovison.preprocessing_yield2crop')
    main()